/*	$Id: plugin.cc 2007/11/12 17:50:00 mallet $ */

/*
 * Copyright (c) 2007 Anthony Mallet                  --  Tue Jun 19 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
static const char id[] = "$Id: plugin.cc 2007/11/12 17:50:00 mallet $";

/* This autotools mess conflicts with omniORB headers. Joy. */
#undef PACKAGE_NAME
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef PACKAGE_STRING
#undef PACKAGE_BUGREPORT

#include <pthread.h>

#include <cstdio>
#include <iostream>
#include <sstream>

#include "Controller/IOserver/include/plugin.h"
#include "bodyinfo.h"
#include "genom-openhrp-plugin.h"
#include "genom-hrp2.h"


/* --- local data ---------------------------------------------------- */

class GenomOpenHRPimpl: public plugin, public POA_GenomOpenHRP {
 private:
  double time;
  double transition;
  motor_command current;
  GenomConfig last, next;

  bool track;
  GenomConfig buffer[(unsigned int)(GHRP2_PLUGIN_BUFFER/GHRP2_PLUGIN_PERIOD)];
  unsigned int read, write;

  int logIndex;
  double log[12000][3];

  pthread_mutex_t mutex;

 public:
  GenomOpenHRPimpl(std::istringstream &strm);
  ~GenomOpenHRPimpl();

  bool setup(robot_state *rs, motor_command *mc);
  void control(robot_state *rs, motor_command *mc);
  bool cleanup(robot_state *rs, motor_command *mc);

  CORBA::Long push(const GenomConfigList &config, CORBA::Double &percent)
    throw (CORBA::SystemException);
  void debug() throw (CORBA::SystemException);
};


/* --- create_plugin ------------------------------------------------- */

plugin *
create_plugin(istringstream &strm)
{
  return new GenomOpenHRPimpl(strm);
}


/* --- GenomOpenHRPimpl::GenomOpenHRPimpl ---------------------------- */

GenomOpenHRPimpl::GenomOpenHRPimpl(std::istringstream &strm)
{
  pthread_mutex_init(&mutex, NULL);

  last.angles.length(DOF);
  last.zmp.length(3);
  last.waistRpy.length(3);

  next.angles.length(DOF);
  next.zmp.length(3);
  next.waistRpy.length(3);
}


/* --- GenomOpenHRPimpl::~GenomOpenHRPimpl --------------------------- */

GenomOpenHRPimpl::~GenomOpenHRPimpl()
{
}


/* --- GenomOpenHRP::setup ------------------------------------------- */

bool
GenomOpenHRPimpl::setup(robot_state *rs, motor_command *mc)
{
  pthread_mutex_lock(&mutex);

  track = false;
  read = write = 0;

  last.time = next.time = time = 0;
  transition = 0.5;
  current = *mc;

  /* copy current config */
  for(unsigned int i = 0; i < DOF; i++)
    last.angles[i] = next.angles[i] = current.angle[i];
  for(unsigned int i = 0; i < 3; i++)
    last.zmp[i] = next.zmp[i] = current.zmp[i];
  for(unsigned int i = 0; i < 3; i++)
    last.waistRpy[i] = next.waistRpy[i] = current.waistRpy[i];

  pthread_mutex_unlock(&mutex);
  return true;
}


/* --- GenomOpenHRP::control ----------------------------------------- */

void
GenomOpenHRPimpl::control(robot_state *rs, motor_command *mc)
{
  if (track) {
    time += GHRP2_PLUGIN_PERIOD;

    /* already have a pending config? */
    if (next.time >= time - GHRP2_PLUGIN_PERIOD/2.) goto interpolate;
  }

  /* rotate last configuration */
  last = next;

  pthread_mutex_lock(&mutex);

  /* empty buffer? */
  if (read == write) {
    pthread_mutex_unlock(&mutex);
    /* stop moving */
    if (track)
      std::cout << "WARNING: config buffer UNDERRUN - STOPPING" << std::endl;
    track = false;
    goto exec;
  }

  /* read new config */
  next = buffer[read];
  if (++read >= sizeof(buffer)/sizeof(buffer[0])) read = 0;

  /* log for debug */
  if (track && logIndex < sizeof(log)/sizeof(log[0])) {
    log[logIndex][0] = time;
    log[logIndex][1] = read;
    log[logIndex][2] = write;
    logIndex++;
  }
  pthread_mutex_unlock(&mutex);

  /* first config? */
  if (!track) {
    last.time = next.time - transition;
    time = last.time + GHRP2_PLUGIN_PERIOD;
    track = true;
    logIndex = 0;
    goto interpolate;
  }

  if (next.time <= last.time) {
    std::cout << "WARNING: next config before previous config - STOPPING" << std::endl;
    track = false;
    goto exec;
  }
  if (next.time < time - GHRP2_PLUGIN_PERIOD/2.) {
    std::cout << "WARNING: config IN THE PAST - STOPPING" << std::endl;
    track = false;
    goto exec;
  }

interpolate: {
    double a = (next.time - time) / (next.time - last.time);

    /* joints */
    for(unsigned int i = 0; i < DOF; i++)
      current.angle[i] = a*last.angles[i] + (1.0-a)*next.angles[i];

    /* zmp */
    for(unsigned int i = 0; i < 3; i++)
      current.zmp[i] = a*last.zmp[i] + (1.0-a)*next.zmp[i];

    /* waist rotations */
    for(unsigned int i = 0; i < 3; i++)
      current.waistRpy[i] = a*last.waistRpy[i] + (1.0-a)*next.waistRpy[i];
  }

exec:
  mc->angle = current.angle;
  mc->zmp = current.zmp;
  mc->waistRpy = current.waistRpy;
}


/* --- GenomOpenHRP::cleanup ----------------------------------------- */

bool
GenomOpenHRPimpl::cleanup(robot_state *rs, motor_command *mc)
{
  track = false;
  return true;
}


/* --- GenomOpenHRP::push -------------------------------------------- */

CORBA::Long
GenomOpenHRPimpl::push(const GenomConfigList &config, CORBA::Double &percent)
  throw (CORBA::SystemException)
{
  unsigned int i;

  for(i = 0; i < config.length(); i++) {
    pthread_mutex_lock(&mutex);

    if (write == read - 1 ||
	(write == sizeof(buffer)/sizeof(buffer[0]) -1 && read == 0)) {
      pthread_mutex_unlock(&mutex);
      break;
    }
    buffer[write] = config[i];
    if (++write >= sizeof(buffer)/sizeof(buffer[0])) write = 0;

    pthread_mutex_unlock(&mutex);
  }

  pthread_mutex_lock(&mutex);
  percent = write - read;
  pthread_mutex_unlock(&mutex);
  percent *= 100./sizeof(buffer)/sizeof(buffer[0]);
  if (percent < 0.) percent += 100.;

  return i;
}


/* --- GenomOpenHRP::debug ------------------------------------------- */

void
GenomOpenHRPimpl::debug() throw (CORBA::SystemException)
{
  FILE *f;
  int i;

  f = fopen("/tmp/plugin-debug", "w");
  for(i = 0; i<logIndex; i++) {
    fprintf(f, "%g %g %g\n", log[i][0], log[i][1], log[i][2]);
  }
  fclose(f);
}
